/*
 * cd_unix.c
 *
 * UNIX(tm) CDROM interface
 */

/* $Id: cd_unix.c,v 1.3 2000/03/15 03:48:04 nyef Exp $ */

/* #define CD_LINUX */		/* use linux-specific code */
#define CD_FREEBSD		/* use freebsd-specific code */
/* #define CD_NETBSD */		/* use netbsd-specific code */

#ifdef CD_LINUX
#define CDROM_DEVICE "/dev/cdrom"
#define CD_HEADER "linux/cdrom.h" /* some idiot #defined linux to 1, so we have to prevent it from being substituted */
#endif

#ifdef CD_FREEBSD
#define CDROM_DEVICE "/dev/acd1c"
#define CD_HEADER <sys/cdio.h>
#define LBA_CONVERT(x) (ntohl(x))
#endif

#ifdef CD_NETBSD
#define CDROM_DEVICE "/dev/rcd0d"
#define CD_HEADER <sys/cdio.h>
#define LBA_CONVERT(x) (x)
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>

#include CD_HEADER

#include "types.h"
#include "cd.h"

int cd_drive;

void cd_set_sector_address(u32 sector)
{
    int result;

    result = lseek(cd_drive, (sector) * 2048, SEEK_SET);
    if (result < 0) {
	perror("cd_unix: seeking to CD sector");
    }
}

void cd_read_next_sector(u8 *buf)
{
    int result;

    result = read(cd_drive, buf, 2048);
    if (result < 0) {
	perror("cd_unix: reading CD data");
    }
}

#ifdef CD_LINUX
void cd_get_tocheader(u8 *start_track, u8 *stop_track)
{
    /* pretend that there are 4 tracks on this CD */
    /* FIXME: real implementation please? */
    *start_track = 1;
    *stop_track = 4;
}

void cd_get_tocentry(u8 track, u32 *lba, u8 *flags)
{
    int result;
    struct cdrom_tocentry cd_tocentry;
    
    cd_tocentry.cdte_track = track;
    cd_tocentry.cdte_format = CDROM_LBA;
    
    result = ioctl(cd_drive, CDROMREADTOCENTRY, &cd_tocentry);
    if (result < 0) {
	perror("cd_unix: reading tocentry");
/*  	deb_printf("cd_unix: tocread failure.\n"); */
    } else {
	*lba = cd_tocentry.cdte_addr.lba;
	*flags = cd_tocentry.cdte_ctrl; /* FIXME: may want cdte.adr << 4 in here as well */
    }
}

#else
void cd_get_tocheader(u8 *start_track, u8 *stop_track)
{
    struct ioc_toc_header toc_header;
    int result;
    
    result = ioctl(cd_drive, CDIOREADTOCHEADER, &toc_header);
    if (result < 0) {
	perror("cd_unix: reading tocheader");
    } else {
	*start_track = toc_header.starting_track;
	*stop_track = toc_header.ending_track;
    }	
}

void cd_get_tocentry(u8 track, u32 *lba, u8 *flags)
{
    int result;
    struct cd_toc_entry toc_entry;
    struct ioc_read_toc_entry cd_tocentry;
    
    cd_tocentry.starting_track = track;
    cd_tocentry.address_format = CD_LBA_FORMAT;
    cd_tocentry.data_len = sizeof(toc_entry);
    cd_tocentry.data = &toc_entry;

    result = ioctl(cd_drive, CDIOREADTOCENTRYS, &cd_tocentry);
    if (result < 0) {
	perror("cd_unix: reading tocentry");
/*  	deb_printf("cd_unix: tocread failure.\n"); */
    } else {
	*lba = LBA_CONVERT(toc_entry.addr.lba);
	*flags = toc_entry.control; /* FIXME: may want addr_type << 4 in here as well */
    }
}
#endif

void cd_init(void)
{
    cd_drive = open(CDROM_DEVICE, O_RDONLY);
    if (cd_drive < 0) {
	perror("cd_unix: opening CD drive");
	cd_drive = 0;
    } else {
	/* Do something? */
    }
}

/*
 * $Log: cd_unix.c,v $
 * Revision 1.3  2000/03/15 03:48:04  nyef
 * cleaned up some comments
 *
 * Revision 1.2  2000/02/20 06:55:35  nyef
 * fixed to hopefully compile on Linux
 *
 * Revision 1.1  2000/02/14 02:09:03  nyef
 * Initial revision
 *
 *
 */
