/*
 * cpu6502.c
 *
 * CAL interface and non-generated code for the cpu6502 core
 */

/* $Id: cpu6502.c,v 1.1 2000/05/01 00:35:33 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include "cal.h"
#include "tool.h"
#include "ui.h"

#include "cpu6502int.h"

void cal_cpu6502_reset(cal_cpu cpu);
void cal_cpu6502_run(cal_cpu cpu);
void cal_cpu6502_runfor(cal_cpu cpu, int cycles);
void cal_cpu6502_irq(cal_cpu cpu, int irqno);
void cal_cpu6502_nmi(cal_cpu cpu);
void cal_cpu6502_setzpage(cal_cpu cpu, void *page0);
int cal_cpu6502_timeleft(cal_cpu cpu);
void cal_cpu6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);

void cal_cpu6502_init(cal_cpu *cpu)
{
    (*cpu)->data.d_cpu6502 = calloc(1, sizeof(struct cpu6502_context));
    if (!(*cpu)->data.d_cpu6502) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_cpu6502_reset;
    (*cpu)->run = cal_cpu6502_run;
    (*cpu)->runfor = cal_cpu6502_runfor;
    (*cpu)->irq = cal_cpu6502_irq;
    (*cpu)->nmi = cal_cpu6502_nmi;
    (*cpu)->setzpage = cal_cpu6502_setzpage;
    (*cpu)->timeleft = cal_cpu6502_timeleft;
    (*cpu)->setmmu8 = cal_cpu6502_setmmu8;
    (*cpu)->data.d_cpu6502->cpu = *cpu;
}

void cal_cpu6502_reset(cal_cpu cpu)
{
    struct cpu6502_context *context;

    context = cpu->data.d_cpu6502;
    context->reg_a = 0x00;
    context->reg_x = 0x00;
    context->reg_y = 0x00;
    context->reg_s = 0xff;

    context->flags = 0x00; /* FIXME: wrong */

    context->pc = context->readfunc(cpu, VECTOR_RESET_LO);
    context->pc |= context->readfunc(cpu, VECTOR_RESET_HI) << 8;
}

void cpu6502_step(struct cpu6502_context *context)
{
    u8 opcode;

    opcode = OPFETCH(context);
    cpu6502_decode[opcode](context);
}

void cal_cpu6502_run(cal_cpu cpu)
{
    struct cpu6502_context *context;

    context = cpu->data.d_cpu6502;

    while (context->cycles_left > 0) {
	cpu6502_step(context);

	if (system_flags & F_UNIMPL) {
	    return;
	}
    }
}

void cal_cpu6502_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_cpu6502->cycles_left += cycles;
}

void cal_cpu6502_irq(cal_cpu cpu, int irqno)
{
    struct cpu6502_context *context;

    context = cpu->data.d_cpu6502;

    if (context->flags & FLAG_I) {
	return; /* interrupts disabled */
    }

    PUSH_BYTE(context, context->pc >> 8);
    PUSH_BYTE(context, context->pc & 0xff);
    PUSH_BYTE(context, context->flags & ~FLAG_B);

    context->flags |= FLAG_I;
    context->flags &= ~FLAG_D;
    
    context->pc = context->readfunc(context->cpu, VECTOR_IRQ_LO);
    context->pc |= context->readfunc(context->cpu, VECTOR_IRQ_HI) << 8;
}

void cal_cpu6502_nmi(cal_cpu cpu)
{
    struct cpu6502_context *context;

    context = cpu->data.d_cpu6502;

    PUSH_BYTE(context, context->pc >> 8);
    PUSH_BYTE(context, context->pc & 0xff);
    PUSH_BYTE(context, context->flags & ~FLAG_B);

    context->pc = context->readfunc(context->cpu, VECTOR_NMI_LO);
    context->pc |= context->readfunc(context->cpu, VECTOR_NMI_HI) << 8;
}

void cal_cpu6502_setzpage(cal_cpu cpu, void *page0)
{
    cpu->data.d_cpu6502->zpage = page0;
}

int cal_cpu6502_timeleft(cal_cpu cpu)
{
    return cpu->data.d_cpu6502->cycles_left;
}

void cal_cpu6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_cpu6502->readfunc = *rtbl;
    cpu->data.d_cpu6502->writefunc = *wtbl;
}

void cpu6502_op_0(struct cpu6502_context *context)
{
    context->pc -= 1;
    deb_printf("unimplemented opcode 0x%02x.\n", OPFETCH(context));
    system_flags |= F_UNIMPL;
}

/*
 * $Log: cpu6502.c,v $
 * Revision 1.1  2000/05/01 00:35:33  nyef
 * Initial revision
 *
 */
