/*
 * event.c
 *
 * periodic event scheduler
 */

/* $Id: event.c,v 1.1 2000/03/10 01:48:36 nyef Exp $ */

#include <stdlib.h>
#include "tool.h"
#include "ui.h"
#include "event.h"

struct event_scheduler {
    struct event *top_event;
    cal_cpu cpu;
};

void event_reschedule(event_scheduler scheduler)
{
    struct event *prev_event;
    struct event *cur_event;
    struct event *next_event;
    int done;

    scheduler->top_event->timeleft = scheduler->top_event->duration;
    prev_event = NULL;
    cur_event = scheduler->top_event;
    next_event = cur_event->next;
    scheduler->top_event = next_event;
    done = 0;

    for (;;) {
	if (!next_event) {
	    done = 1;
	} else if (cur_event->timeleft <= next_event->timeleft) {
	    next_event->timeleft -= cur_event->timeleft;
	    done = 1;
	}
	
	if (done) {
	    if (prev_event) {
		prev_event->next = cur_event;
	    } else {
		scheduler->top_event = cur_event;
	    }
	    cur_event->next = next_event;
	    return;
	}
	
	cur_event->timeleft -= next_event->timeleft;
	
	prev_event = next_event;
	next_event = next_event->next;
    }
}

void event_register(event_scheduler scheduler, struct event *new_event)
{
    new_event->next = scheduler->top_event;
    scheduler->top_event = new_event;
    event_reschedule(scheduler);
}

void event_timeslice(event_scheduler scheduler)
{
    for (;;) {
	event_reschedule(scheduler);

	scheduler->cpu->runfor(scheduler->cpu, scheduler->top_event->timeleft);
	scheduler->cpu->run(scheduler->cpu);
	
	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
	    system_flags |= F_QUIT;
	    return;
	}
	
	if (!scheduler->top_event->callback) {
	    return;
	}
	scheduler->top_event->callback();
    }
}

event_scheduler new_event_scheduler(cal_cpu cpu)
{
    event_scheduler retval;

    retval = malloc(sizeof(struct event_scheduler));

    if (!retval) {
	return NULL;
    }

    retval->cpu = cpu;
    retval->top_event = NULL;

    return retval;
}

/*
 * $Log: event.c,v $
 * Revision 1.1  2000/03/10 01:48:36  nyef
 * Initial revision
 *
 */
